# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from spectre_core.spectrograms.spectrogram import Spectrogram
from spectre_core.plotting.base import BasePanel
from spectre_core.plotting.panel_register import panels
from spectre_core.exceptions import PanelNotFoundError

def get_panel(panel_name: str, 
              spectrogram: Spectrogram,
              time_type: str,
              *args, 
              **kwargs) -> BasePanel:
   
    Panel = panels.get(panel_name)
    if Panel is None:
        valid_panels = list(panels.keys())
        raise PanelNotFoundError(f"Could not find panel with name {panel_name}. "
                                 f"Expected one of {valid_panels}")
    
    return Panel(panel_name,
                 spectrogram, 
                 time_type,
                 *args, 
                 **kwargs)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from typing import Optional, Any
from warnings import warn
from datetime import datetime, timedelta
from dataclasses import dataclass

import numpy as np
from astropy.io import fits

from spectre_core.file_handlers.configs import FitsConfig
from spectre_core.cfg import DEFAULT_DATETIME_FORMAT, get_chunks_dir_path
from spectre_core.spectrograms.array_operations import (
    find_closest_index,
    normalise_peak_intensity,
    compute_resolution,
    compute_range,
    subtract_background,
)

@dataclass
class FrequencyCut:
    time: float | datetime
    frequencies: np.ndarray
    cut: np.ndarray
    spectrum_type: str


@dataclass
class TimeCut:
    frequency: float
    times: np.ndarray
    cut: np.ndarray
    spectrum_type: str
    

class Spectrogram:
    def __init__(self, 
                 dynamic_spectra: np.ndarray, # holds the spectrogram data
                 times: np.ndarray, # holds the time stamp [s] for each spectrum
                 frequencies: np.ndarray,  # physical frequencies [Hz] for each spectral component
                 tag: str,
                 chunk_start_time: Optional[str] = None, 
                 microsecond_correction: int = 0, 
                 spectrum_type: Optional[str] = None,
                 start_background: Optional[str] = None, 
                 end_background: Optional[str] = None): 
        
        # dynamic spectra
        self._dynamic_spectra = dynamic_spectra
        self._dynamic_spectra_as_dBb: Optional[np.ndarray] = None # cache

        # assigned times and frequencies
        self._times = times
        self._datetimes: Optional[list[datetime]] = None # cache
        self._frequencies = frequencies

        # general metadata
        self._tag = tag
        self._chunk_start_time = chunk_start_time
        self._chunk_start_datetime: Optional[datetime] = None # cache
        self._microsecond_correction = microsecond_correction
        self._spectrum_type = spectrum_type

        # background metadata     
        self._background_spectrum: Optional[np.ndarray] = None # cache
        self._start_background = start_background
        self._end_background = end_background
        self._start_background_index = 0 # by default
        self._end_background_index = self.num_times # by default
        self._check_shapes()


    @property
    def dynamic_spectra(self) -> np.ndarray:
        return self._dynamic_spectra
    

    @property
    def times(self) -> np.ndarray:
        return self._times
    
    
    @property
    def num_times(self) -> int:
        return len(self._times)
    

    @property
    def time_resolution(self) -> float:
        return compute_resolution(self._times)
    

    @property
    def time_range(self) -> float:
        return compute_range(self._times)
    

    @property
    def datetimes(self) -> list[datetime]:
        if self._datetimes is None:
            self._datetimes = [self.chunk_start_datetime + timedelta(seconds=(t + self.microsecond_correction*1e-6)) for t in self._times]
        return self._datetimes
    

    @property
    def frequencies(self) -> np.ndarray:
        return self._frequencies


    @property
    def num_frequencies(self) -> int:
        return len(self._frequencies)
    
    
    @property
    def frequency_resolution(self) -> float:
        return compute_resolution(self._frequencies)
    

    @property
    def frequency_range(self) -> float:
        return compute_range(self._frequencies)
    

    @property
    def tag(self) -> str:
        return self._tag
    

    @property
    def chunk_start_time(self) -> str:
        if self._chunk_start_time is None:
            raise AttributeError(f"Chunk start time has not been set.")
        return self._chunk_start_time
    

    @property
    def chunk_start_datetime(self) -> datetime:
        if self._chunk_start_datetime is None:
            self._chunk_start_datetime = datetime.strptime(self.chunk_start_time, DEFAULT_DATETIME_FORMAT)
        return self._chunk_start_datetime


    @property
    def microsecond_correction(self) -> int:
        return self._microsecond_correction
    

    @property
    def spectrum_type(self) -> Optional[str]:
        return self._spectrum_type
    

    @property
    def start_background(self) -> Optional[str]:
        return self._start_background
    

    @property
    def end_background(self) -> Optional[str]:
        return self._end_background
    
    
    @property
    def background_spectrum(self) -> np.ndarray:
        if self._background_spectrum is None:
            self._background_spectrum = np.nanmean(self._dynamic_spectra[:, self._start_background_index:self._end_background_index+1], 
                                               axis=-1)
        return self._background_spectrum
    

    @property
    def dynamic_spectra_as_dBb(self) -> np.ndarray:
        if self._dynamic_spectra_as_dBb is None:
            # Create an artificial spectrogram where each spectrum is identically the background spectrum
            background_spectra = self.background_spectrum[:, np.newaxis]
            # Suppress divide by zero and invalid value warnings for this block of code
            with np.errstate(divide='ignore'):
                # Depending on the spectrum type, compute the dBb values differently
                if self._spectrum_type == "amplitude" or self._spectrum_type == "digits":
                    self._dynamic_spectra_as_dBb = 10 * np.log10(self._dynamic_spectra / background_spectra)
                elif self._spectrum_type == "power":
                    self._dynamic_spectra_as_dBb = 20 * np.log10(self._dynamic_spectra / background_spectra)
                else:
                    raise NotImplementedError(f"{self.spectrum_type} unrecognised, uncertain decibel conversion!")
        return self._dynamic_spectra_as_dBb  
    
    
    def set_background(self, 
                       start_background: str, 
                       end_background: str) -> None:
        """Public setter for start and end of the background"""
        self._dynamic_spectra_as_dBb = None # reset cache
        self._background_spectrum = None # reset cache
        self._start_background = start_background
        self._end_background = end_background
        self._update_background_indices_from_interval()
    
    
    
    def _update_background_indices_from_interval(self) -> None:
        start_background = datetime.strptime(self._start_background, DEFAULT_DATETIME_FORMAT)
        self._start_background_index = find_closest_index(start_background, 
                                                          self.datetimes, 
                                                          enforce_strict_bounds=True)
        
        end_background = datetime.strptime(self._end_background, DEFAULT_DATETIME_FORMAT)
        self._end_background_index = find_closest_index(end_background, 
                                                        self.datetimes, 
                                                        enforce_strict_bounds=True)


    def _check_shapes(self) -> None:
        num_spectrogram_dims = np.ndim(self._dynamic_spectra)
        # Check if 'dynamic_spectra' is a 2D array
        if num_spectrogram_dims != 2:
            raise ValueError(f"Expected dynamic spectrogram to be a 2D array, but got {num_spectrogram_dims}D array")
        dynamic_spectra_shape = self.dynamic_spectra.shape
        # Check if the dimensions of 'dynamic_spectra' are consistent with the time and frequency arrays
        if dynamic_spectra_shape[0] != self.num_frequencies:
            raise ValueError(f"Mismatch in number of frequency bins: Expected {self.num_frequencies}, but got {dynamic_spectra_shape[0]}")
        
        if dynamic_spectra_shape[1] != self.num_times:
            raise ValueError(f"Mismatch in number of time bins: Expected {self.num_times}, but got {dynamic_spectra_shape[1]}")
        

    def save(self) -> None:
        fits_config = FitsConfig(self._tag)
        fits_config = fits_config if fits_config.exists else {}

        chunk_start_datetime = self.chunk_start_datetime
        chunk_parent_path = get_chunks_dir_path(year = chunk_start_datetime.year,
                                                month = chunk_start_datetime.month,
                                                day = chunk_start_datetime.day)
        file_name = f"{self.chunk_start_time}_{self._tag}.fits"
        write_path = os.path.join(chunk_parent_path, file_name)
        _save_spectrogram(write_path, self, fits_config)
    

    def integrate_over_frequency(self, 
                                 correct_background: bool = False, 
                                 peak_normalise: bool = False) -> np.ndarray[np.float32]:
        
        # integrate over frequency
        I = np.trapz(self._dynamic_spectra, self._frequencies, axis=0)

        if correct_background:
            I = subtract_background(I, 
                                    self._start_background_index, 
                                    self._end_background_index)
        if peak_normalise:
            I = normalise_peak_intensity(I)
        return I


    def get_frequency_cut(self, 
                          at_time: float | str,
                          dBb: bool = False,
                          peak_normalise: bool = False) -> FrequencyCut:
        
        # it is important to note that the "at time" specified by the user likely does not correspond
        # exactly to one of the times assigned to each spectrogram. So, we compute the nearest achievable,
        # and return it from the function as output too.
        if isinstance(at_time, str):
            at_time = datetime.strptime(at_time, DEFAULT_DATETIME_FORMAT)
            index_of_cut = find_closest_index(at_time, 
                                              self.datetimes, 
                                              enforce_strict_bounds = True)
            time_of_cut = self.datetimes[index_of_cut]  

        elif isinstance(at_time, (float, int)):
            index_of_cut = find_closest_index(at_time, 
                                              self._times, 
                                              enforce_strict_bounds = True)
            time_of_cut = self.times[index_of_cut]
        
        else:
            raise ValueError(f"Type of at_time is unsupported: {type(at_time)}")
        
        if dBb:
            ds = self.dynamic_spectra_as_dBb
        else:
            ds = self._dynamic_spectra
        
        cut = ds[:, index_of_cut].copy() # make a copy so to preserve the spectrum on transformations of the cut

        if dBb:
            if peak_normalise:
                warn("Ignoring frequency cut normalisation, since dBb units have been specified")
        else:
            if peak_normalise:
                cut = normalise_peak_intensity(cut)
        
        return FrequencyCut(time_of_cut, 
                            self._frequencies, 
                            cut, 
                            self._spectrum_type)

        
    def get_time_cut(self,
                     at_frequency: float,
                     dBb: bool = False,
                     peak_normalise = False, 
                     correct_background = False, 
                     return_time_type: str = "seconds") -> TimeCut:
        
        # it is important to note that the "at frequency" specified by the user likely does not correspond
        # exactly to one of the physical frequencies assigned to each spectral component. So, we compute the nearest achievable,
        # and return it from the function as output too.
        index_of_cut = find_closest_index(at_frequency, 
                                          self._frequencies, 
                                          enforce_strict_bounds = True)
        frequency_of_cut = self.frequencies[index_of_cut]

        if return_time_type == "datetimes":
            times = self.datetimes
        elif return_time_type == "seconds":
            times = self.times
        else:
            raise ValueError(f"Invalid return_time_type. Got {return_time_type}, expected one of 'datetimes' or 'seconds'")

        # dependent on the requested cut type, we return the dynamic spectra in the preferred units
        if dBb:
            ds = self.dynamic_spectra_as_dBb
        else:
            ds = self.dynamic_spectra
        
        cut = ds[index_of_cut,:].copy() # make a copy so to preserve the spectrum on transformations of the cut

        # Warn if dBb is used with background correction or peak normalisation
        if dBb:
            if correct_background or peak_normalise:
                warn("Ignoring time cut normalisation, since dBb units have been specified")
        else:
            # Apply background correction if required
            if correct_background:
                cut = subtract_background(cut, 
                                          self._start_background_index,
                                          self._end_background_index)
            
            # Apply peak normalisation if required
            if peak_normalise:
                cut = normalise_peak_intensity(cut)

        return TimeCut(frequency_of_cut, 
                         times, 
                         cut,
                         self.spectrum_type)
    

def _seconds_of_day(dt: datetime) -> float:
    start_of_day = datetime(dt.year, dt.month, dt.day)
    return (dt - start_of_day).total_seconds()


# Function to create a FITS file with the specified structure
def _save_spectrogram(write_path: str, 
                      spectrogram: Spectrogram, 
                      fits_config: FitsConfig | dict[str, Any]) -> None:
    if spectrogram.chunk_start_time is None:
        raise ValueError(f"Spectrogram must have a defined chunk_start_time. Received {spectrogram.chunk_start_time}")
    
    # Primary HDU with data
    primary_data = spectrogram.dynamic_spectra.astype(dtype=np.float32) 
    primary_hdu = fits.PrimaryHDU(primary_data)

    primary_hdu.header.set('SIMPLE', True, 'file does conform to FITS standard')
    primary_hdu.header.set('BITPIX', -32, 'number of bits per data pixel')
    primary_hdu.header.set('NAXIS', 2, 'number of data axes')
    primary_hdu.header.set('NAXIS1', spectrogram.dynamic_spectra.shape[1], 'length of data axis 1')
    primary_hdu.header.set('NAXIS2', spectrogram.dynamic_spectra.shape[0], 'length of data axis 2')
    primary_hdu.header.set('EXTEND', True, 'FITS dataset may contain extensions')

    # Add comments
    comments = [
        "FITS (Flexible Image Transport System) format defined in Astronomy and",
        "Astrophysics Supplement Series v44/p363, v44/p371, v73/p359, v73/p365.",
        "Contact the NASA Science Office of Standards and Technology for the",
        "FITS Definition document #100 and other FITS information."
    ]
    
    # The comments section remains unchanged since add_comment is the correct approach
    for comment in comments:
        primary_hdu.header.add_comment(comment)

    start_datetime = spectrogram.datetimes[0]
    start_date = start_datetime.strftime("%Y-%m-%d")
    start_time = start_datetime.strftime("%H:%M:%S.%f")

    end_datetime = spectrogram.datetimes[-1]
    end_date = end_datetime.strftime("%Y-%m-%d")
    end_time = end_datetime.strftime("%H:%M:%S.%f")

    primary_hdu.header.set('DATE', start_date, 'time of observation')
    primary_hdu.header.set('CONTENT', f'{start_date} dynamic spectrogram', 'title of image')
    primary_hdu.header.set('ORIGIN', f'{fits_config.get("ORIGIN")}')
    primary_hdu.header.set('TELESCOP', f'{fits_config.get("TELESCOP")} tag: {spectrogram.tag}', 'type of instrument')
    primary_hdu.header.set('INSTRUME', f'{fits_config.get("INSTRUME")}') 
    primary_hdu.header.set('OBJECT', f'{fits_config.get("OBJECT")}', 'object description')

    primary_hdu.header.set('DATE-OBS', f'{start_date}', 'date observation starts')
    primary_hdu.header.set('TIME-OBS', f'{start_time}', 'time observation starts')
    primary_hdu.header.set('DATE-END', f'{end_date}', 'date observation ends')
    primary_hdu.header.set('TIME-END', f'{end_time}', 'time observation ends')

    primary_hdu.header.set('BZERO', 0, 'scaling offset')
    primary_hdu.header.set('BSCALE', 1, 'scaling factor')
    primary_hdu.header.set('BUNIT', f"{spectrogram.spectrum_type}", 'z-axis title') 

    primary_hdu.header.set('DATAMIN', np.nanmin(spectrogram.dynamic_spectra), 'minimum element in image')
    primary_hdu.header.set('DATAMAX', np.nanmax(spectrogram.dynamic_spectra), 'maximum element in image')

    primary_hdu.header.set('CRVAL1', f'{_seconds_of_day(start_datetime)}', 'value on axis 1 at reference pixel [sec of day]')
    primary_hdu.header.set('CRPIX1', 0, 'reference pixel of axis 1')
    primary_hdu.header.set('CTYPE1', 'TIME [UT]', 'title of axis 1')
    primary_hdu.header.set('CDELT1', spectrogram.time_resolution, 'step between first and second element in x-axis')

    primary_hdu.header.set('CRVAL2', 0, 'value on axis 2 at reference pixel')
    primary_hdu.header.set('CRPIX2', 0, 'reference pixel of axis 2')
    primary_hdu.header.set('CTYPE2', 'Frequency [Hz]', 'title of axis 2')
    primary_hdu.header.set('CDELT2', spectrogram.frequency_resolution, 'step between first and second element in axis')

    primary_hdu.header.set('OBS_LAT', f'{fits_config.get("OBS_LAT")}', 'observatory latitude in degree')
    primary_hdu.header.set('OBS_LAC', 'N', 'observatory latitude code {N,S}')
    primary_hdu.header.set('OBS_LON', f'{fits_config.get("OBS_LON")}', 'observatory longitude in degree')
    primary_hdu.header.set('OBS_LOC', 'W', 'observatory longitude code {E,W}')
    primary_hdu.header.set('OBS_ALT', f'{fits_config.get("OBS_ALT")}', 'observatory altitude in meter asl')


    # Wrap arrays in an additional dimension to mimic the e-CALLISTO storage
    times_wrapped = np.array([spectrogram.times.astype(np.float32)])
    # To mimic e-Callisto storage, convert frequencies to MHz
    frequencies_MHz = spectrogram.frequencies *1e-6
    frequencies_wrapped = np.array([frequencies_MHz.astype(np.float32)])
    
    # Binary Table HDU (extension)
    col1 = fits.Column(name='TIME', format='PD', array=times_wrapped)
    col2 = fits.Column(name='FREQUENCY', format='PD', array=frequencies_wrapped)
    cols = fits.ColDefs([col1, col2])

    bin_table_hdu = fits.BinTableHDU.from_columns(cols)

    bin_table_hdu.header.set('PCOUNT', 0, 'size of special data area')
    bin_table_hdu.header.set('GCOUNT', 1, 'one data group (required keyword)')
    bin_table_hdu.header.set('TFIELDS', 2, 'number of fields in each row')
    bin_table_hdu.header.set('TTYPE1', 'TIME', 'label for field 1')
    bin_table_hdu.header.set('TFORM1', 'D', 'data format of field: 8-byte DOUBLE')
    bin_table_hdu.header.set('TTYPE2', 'FREQUENCY', 'label for field 2')
    bin_table_hdu.header.set('TFORM2', 'D', 'data format of field: 8-byte DOUBLE')
    bin_table_hdu.header.set('TSCAL1', 1, '')
    bin_table_hdu.header.set('TZERO1', 0, '')
    bin_table_hdu.header.set('TSCAL2', 1, '')
    bin_table_hdu.header.set('TZERO2', 0, '')

    # Create HDU list and write to file
    hdul = fits.HDUList([primary_hdu, bin_table_hdu])
    hdul.writeto(write_path, overwrite=True)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any
import json

from spectre_core.file_handlers.base import BaseFileHandler

class JsonHandler(BaseFileHandler):
    def __init__(self, 
                 parent_path: str, 
                 base_file_name: str,
                 extension: str = "json",
                 **kwargs):
        super().__init__(parent_path, 
                         base_file_name, 
                         extension,
                         **kwargs)
    
    
    def read(self) -> dict[str, Any]:
        with open(self.file_path, 'r') as f:
            return json.load(f)
        

    def save(self, 
             d: dict, 
             doublecheck_overwrite: bool = True) -> None:
        self.make_parent_path()

        if self.exists and doublecheck_overwrite:
            self.doublecheck_overwrite()
        
        with open(self.file_path, 'w') as file:
            json.dump(d, file, indent=4)


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, Optional, Type, Tuple
from abc import ABC
import ast

from spectre_core.file_handlers.json import JsonHandler
from spectre_core.cfg import JSON_CONFIGS_DIR_PATH
from spectre_core.exceptions import InvalidTagError


def _unpack_param(param: str) -> list[str, str]:
    """Seperate a string of the form "a=b" into a list [a,b]."""
    if not param or '=' not in param:
        raise ValueError(f'Invalid format: "{param}". Expected "KEY=VALUE".')
    if param.startswith('=') or param.endswith('='):
        raise ValueError(f'Invalid format: "{param}". Expected "KEY=VALUE".')
    # remove leading and trailing whitespace.
    param = param.strip()
    return param.split('=', 1)


def _params_to_string_dict(params: list[str]) -> dict[str, str]:
    """Converts a list with string elements of the form "a=b" into a dictionary where the key value pairs are "a": "b"."""
    d = {}
    for param in params:
        key, value = _unpack_param(param)
        d[key] = value
    return d


def _convert_to_dict(v: str) -> dict:
    """Evaluate literally a string containing a Python dictionary expression."""
    return ast.literal_eval(v)


def _convert_to_bool(v: str) -> bool:
    """Evaluate literally a string representation of a boolean as a boolean"""
    v = v.lower()
    if v in ('true', '1', 't', 'y', 'yes'):
        return True
    if v in ('false', '0', 'f', 'n', 'no'):
        return False
    raise ValueError(f'Cannot convert {v} to bool.')
    

def _convert_string_to_type(value: str, 
                           target_type: Type) -> Any:
    """Cast a string as the target type."""
    if target_type == bool:
        return _convert_to_bool(value)
    elif target_type == dict:
        return _convert_to_dict(value)
    return target_type(value)


def _type_cast_string_dict(d: dict[str, str], 
                           type_template: dict[str, Type]) -> dict[str, Any]:
    """Cast the values of the input dictionary according to a type template."""
    casted_d = {}
    for key, value in d.items():
        target_type = type_template.get(key)
        if target_type is None:
            raise KeyError(f'Key "{key}" not found in type template. Expected keys: {list(type_template.keys())}')
        try:
            casted_d[key] = _convert_string_to_type(value, target_type)
        except ValueError:
            raise ValueError(f'Failed to convert key "{key}" with value "{value}" to {target_type.__name__}.')
    return casted_d


def _validate_keys(d: dict[str, Any], 
                   type_template: dict[str, type], 
                   ignore_keys: Optional[list] = None) -> None:
    """Validate that the keys in the input dictionary map one-to-one to the input type template."""
    if ignore_keys is None:
        ignore_keys = []

    type_template_keys = set(type_template.keys())
    input_keys = set(d.keys())
    ignore_keys = set(ignore_keys)

    missing_keys = type_template_keys - input_keys
    invalid_keys = input_keys - type_template_keys - ignore_keys

    errors = []
    
    if missing_keys:
        errors.append(f"Missing keys: {', '.join(missing_keys)}")

    if invalid_keys:
        errors.append(f"Invalid keys: {', '.join(invalid_keys)}")
    
    if errors:
        raise KeyError("Key errors found! " + " ".join(errors))


def _validate_types(d: dict[str, Any], 
                    type_template: dict[str, type], 
                    ignore_keys: Optional[list] = None) -> None:
    """Validate the types in the input dictionary are consistent with the input type template."""

    if ignore_keys is None:
        ignore_keys = []
        
    for k, v in d.items():
        if k in ignore_keys:
            continue
        expected_type = type_template[k]
        if expected_type is None:
            raise KeyError(f'Type not found for key "{k}" in the type template.')
        
        if not isinstance(v, expected_type):
            raise TypeError(f'Expected {expected_type} for "{k}", but got {type(v)}.')


def validate_against_type_template(d: dict[str, Any], 
                                   type_template: dict[str, type], 
                                   ignore_keys: Optional[list] = None) -> None:
    """Validates the keys and values of the input dictionary, according to the input type template."""
    _validate_keys(d, 
                   type_template, 
                   ignore_keys = ignore_keys)
    _validate_types(d, 
                    type_template, 
                    ignore_keys = ignore_keys)


def type_cast_params(params: list[str], 
                     type_template: dict[str, type]) -> dict[str, Any]:
    d = _params_to_string_dict(params)
    return _type_cast_string_dict(d, 
                                  type_template)



class SPECTREConfig(JsonHandler, ABC):
    def __init__(self, 
                 tag: str, 
                 config_type: str, 
                 **kwargs):
        self._validate_tag(tag)
        self._tag = tag
        self._config_type = config_type

        self._dict = None # cache
        super().__init__(JSON_CONFIGS_DIR_PATH, 
                         f"{config_type}_config_{tag}", 
                         **kwargs)


    @property
    def tag(self) -> str:
        return self._tag
    

    @property
    def config_type(self) -> str:
        return self._config_type
    

    @property
    def dict(self) -> dict[str, Any]:
        if self._dict is None:
            self._dict = self.read()
        return self._dict
    

    def _validate_tag(self, tag: str) -> None:
        if "_" in tag:
            raise InvalidTagError(f"Tags cannot contain an underscore. Received {tag}")
        if "callisto" in tag:
            raise InvalidTagError(f'"callisto" cannot be a substring in a native tag. Received "{tag}"')

    
    def __getitem__(self, 
                    key: str) -> Any:
        return self.dict[key]
    

    def get(self, 
            *args, 
            **kwargs) -> Any:
        return self.dict.get(*args, 
                             **kwargs)
    
    
    def update(self, 
               *args, 
               **kwargs) -> None:
        self.dict.update(*args, **kwargs)

    
    def items(self):
        return self.dict.items()
    
    
    def keys(self):
        return self.dict.keys()
    
    
    def values(self):
        return self.dict.values()


class FitsConfig(SPECTREConfig):

    type_template = {
        "ORIGIN": str,
        "TELESCOP": str,
        "INSTRUME": str,
        "OBJECT": str,
        "OBS_LAT": float,
        "OBS_LONG": float,
        "OBS_ALT": float
    }

    def __init__(self, 
                 tag: str, 
                 **kwargs):
        super().__init__(tag, 
                         "fits", 
                         **kwargs)

    def get_create_fits_config_cmd(self, 
                                   tag: str, 
                                   as_string: bool = False) -> list[str] | str:
        command_as_list = ["spectre", "create", "fits-config", "-t", tag]
        for key, value in self.type_template.items():
            command_as_list += ["-p"]
            command_as_list += [f"{key}={value.__name__}"]
        if as_string:
            return " ".join(command_as_list)
        else:
            return command_as_list
        

    def save_params(self, 
                    params: list[str], 
                    doublecheck_overwrite: bool = True
                    ) -> None:
        d = type_cast_params(params, 
                             self.type_template)
        self.save(d, 
                  doublecheck_overwrite = doublecheck_overwrite)
        
    
class CaptureConfig(SPECTREConfig):
    def __init__(self, 
                 tag: str, 
                 **kwargs):
        super().__init__(tag, 
                         "capture", 
                         **kwargs)
        

    def get_receiver_metadata(self) -> Tuple[str, str]:

        receiver_name, mode = self.get("receiver"), self.get("mode")

        if receiver_name is None:
            raise ValueError("Invalid capture config! Receiver name is not specified.")
    
        if mode is None:
            raise ValueError("Invalid capture config! Receiver mode is not specified.")
        
        return receiver_name, mode


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import subprocess
import shutil
import gzip
from datetime import datetime
from typing import Optional

from spectre_core.cfg import (
    DEFAULT_DATETIME_FORMAT,
    CALLISTO_INSTRUMENT_CODES
)
from spectre_core.cfg import get_chunks_dir_path

temp_dir = os.path.join(os.environ['SPECTRE_DIR_PATH'], "tmp")

def get_chunk_name(station: str, date: str, time: str, instrument_code: str) -> str:
    dt = datetime.strptime(f"{date}T{time}", '%Y%m%dT%H%M%S')
    formatted_time = dt.strftime(DEFAULT_DATETIME_FORMAT)
    return f"{formatted_time}_callisto-{station.lower()}-{instrument_code}.fits"


def get_chunk_components(gz_path: str):
    file_name = os.path.basename(gz_path)
    if not file_name.endswith(".fit.gz"):
        raise ValueError(f"Unexpected file extension in {file_name}. Expected .fit.gz")
    
    file_base_name = file_name.rstrip(".fit.gz")
    parts = file_base_name.split('_')
    if len(parts) != 4:
        raise ValueError("Filename does not conform to the expected format of [station]_[date]_[time]_[instrument_code]")
    
    return parts


def get_chunk_path(gz_path: str) -> str:
    station, date, time, instrument_code = get_chunk_components(gz_path)
    fits_chunk_name = get_chunk_name(station, date, time, instrument_code)
    chunk_start_time = fits_chunk_name.split('_')[0]
    chunk_start_datetime = datetime.strptime(chunk_start_time, DEFAULT_DATETIME_FORMAT)
    chunk_parent_path = get_chunks_dir_path(year = chunk_start_datetime.year,
                                            month = chunk_start_datetime.month,
                                            day = chunk_start_datetime.day)
    if not os.path.exists(chunk_parent_path):
        os.makedirs(chunk_parent_path)
    return os.path.join(chunk_parent_path, fits_chunk_name)


def unzip_file_to_chunks(gz_path: str):
    fits_path = get_chunk_path(gz_path)
    with gzip.open(gz_path, 'rb') as f_in, open(fits_path, 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)


def unzip_to_chunks():
    for entry in os.scandir(temp_dir):
        if entry.is_file() and entry.name.endswith('.gz'):
            unzip_file_to_chunks(entry.path)
            os.remove(entry.path)


def download_callisto_data(instrument_code: str, 
                           year: int, 
                           month: int, 
                           day: int):
    date_str = f"{year:04d}/{month:02d}/{day:02d}"
    base_url = f"http://soleil.i4ds.ch/solarradio/data/2002-20yy_Callisto/{date_str}/"
    command = [
        'wget', '-r', '-l1', '-nd', '-np', 
        '-R', '.tmp',
        '-A', f'{instrument_code}*.fit.gz',
        '-P', temp_dir,
        base_url
    ]

    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        print(f"An error occurred: {e}")

def fetch_chunks(instrument_code: Optional[str], 
                 year: Optional[int], 
                 month: Optional[int], 
                 day: Optional[int]):


    if (year is None) or (month is None) or (day is None):
        raise ValueError(f"All of year, month and day should be specified")
    
    if not os.path.exists(temp_dir):
        os.mkdir(temp_dir)

    if instrument_code not in CALLISTO_INSTRUMENT_CODES:
        raise ValueError(f"No match found for \"{instrument_code}\". Expected one of {CALLISTO_INSTRUMENT_CODES}")

    download_callisto_data(instrument_code, year, month, day)
    unzip_to_chunks()
    shutil.rmtree(temp_dir)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Tuple
from typing import Optional
from datetime import timedelta
from dataclasses import dataclass

import numpy as np

from spectre_core.chunks.chunk_register import register_chunk
from spectre_core.spectrograms.spectrogram import Spectrogram
from spectre_core.chunks.library.fixed.chunk import BinChunk, FitsChunk
from spectre_core.cfg import DEFAULT_DATETIME_FORMAT
from spectre_core.chunks.base import SPECTREChunk, ChunkFile
from spectre_core.exceptions import InvalidSweepMetadataError


@dataclass
class SweepMetadata:
    """Wrapper for metadata required to assign center frequencies to each IQ sample in the chunk.
    
    center_frequencies is an ordered list containing all the center frequencies that the IQ samples
    were collected at. Typically, these will be ordered in "steps", where each step corresponds to
    IQ samples collected at a constant center frequency:

    (freq_0, freq_1, ..., freq_M, freq_0, freq_1, ..., freq_M, ...), freq_0 < freq_1 < ... < freq_M

    The n'th element of the num_samples list, tells us how many samples were collected at the n'th
    element of center_frequencies:

    Number of samples: (num_samples_at_freq_0, num_samples_at_freq_1, ...)

    Both these lists together allow us to map for each IQ sample, the center frequency it was collected at.
    """
    center_frequencies: np.ndarray
    num_samples: np.ndarray


@register_chunk('sweep')
class Chunk(SPECTREChunk):
    def __init__(self, chunk_start_time, tag):
        super().__init__(chunk_start_time, tag)

        self.add_file(BinChunk(self.chunk_parent_path, self.chunk_name))
        self.add_file(FitsChunk(self.chunk_parent_path, self.chunk_name))
        self.add_file(HdrChunk(self.chunk_parent_path, self.chunk_name))


    def build_spectrogram(self, 
                          previous_chunk: Optional[SPECTREChunk] = None) -> Spectrogram:
        """Create a spectrogram by performing a Short Time FFT on the (swept) IQ samples for this chunk."""
        iq_data = self.read_file("bin")
        millisecond_correction, sweep_metadata = self.read_file("hdr")

        if previous_chunk:
            iq_data, sweep_metadata, num_samples_prepended = self.__reconstruct_initial_sweep(previous_chunk,
                                                                                              iq_data,
                                                                                              sweep_metadata)
            
            chunk_start_time, millisecond_correction = self.__correct_timing(millisecond_correction,
                                                                             num_samples_prepended)
        else:
            chunk_start_time = self.chunk_start_time

        microsecond_correction = millisecond_correction * 1e3

        times, frequencies, dynamic_spectra = self.__do_STFFT(iq_data, 
                                                              sweep_metadata)
        
        return Spectrogram(dynamic_spectra,
                           times,
                           frequencies,
                           self.tag,
                           chunk_start_time,
                           microsecond_correction,
                           spectrum_type = "amplitude")
    
    def __get_final_sweep(self,
                          previous_chunk: SPECTREChunk) -> Tuple[np.ndarray, SweepMetadata]:
        """Get data from the final sweep of the previous chunk."""
        # unpack the data from the previous chunk
        previous_iq_data = previous_chunk.read_file("bin")
        _, previous_sweep_metadata = previous_chunk.read_file("hdr")
        # find the step index from the last sweep
        # [0] since the return of np.where is a 1 element Tuple, 
        # containing a list of step indices corresponding to the smallest center frequencies
        # [-1] since we want the final step index, where the center frequency is minimised
        final_sweep_start_step_index = np.where(previous_sweep_metadata.center_frequencies == np.min(previous_sweep_metadata.center_frequencies))[0][-1]
        # isolate the data from the final sweep
        final_center_frequencies = previous_sweep_metadata.center_frequencies[final_sweep_start_step_index:]
        final_num_samples = previous_sweep_metadata.num_samples[final_sweep_start_step_index:]
        final_sweep_iq_data = previous_iq_data[-np.sum(final_num_samples):]

        # sanity check on the number of samples in the final sweep
        if len(final_sweep_iq_data) != np.sum(final_num_samples):
            raise ValueError((f"Unexpected error! Mismatch in sample count for the final sweep data."
                              f"Expected {np.sum(final_num_samples)} based on sweep metadata, but found "
                              f" {len(final_sweep_iq_data)} IQ samples in the final sweep"))
        
        return final_sweep_iq_data, SweepMetadata(final_center_frequencies, final_num_samples)


    def __prepend_iq_data(self,
                          carryover_iq_data: np.ndarray,
                          iq_data: np.ndarray) -> np.ndarray:
        """Prepend the IQ samples from the final sweep of the previous chunk."""
        return np.concatenate((carryover_iq_data, iq_data))
    

    def __prepend_center_frequencies(self,
                                     carryover_center_frequencies: np.ndarray,
                                     center_frequencies: np.ndarray,
                                     final_sweep_spans_two_chunks: bool)-> np.ndarray:
        """Prepend the center frequencies from the final sweep of the previous chunk."""
        # in the case that the sweep has bled across chunks,
        # do not permit identical neighbours in the center frequency array
        if final_sweep_spans_two_chunks:
            # truncate the final frequency to prepend (as it already exists in the array we are appending to in this case)
            carryover_center_frequencies = carryover_center_frequencies[:-1]
        return np.concatenate((carryover_center_frequencies, center_frequencies))
    

    def __prepend_num_samples(self,
                              carryover_num_samples: np.ndarray,
                              num_samples: np.ndarray,
                              final_sweep_spans_two_chunks: bool) -> np.ndarray:
        """Prepend the number of samples from the final sweep of the previous chunk."""
        if final_sweep_spans_two_chunks:
            # ensure the number of samples from the final step in the previous chunk are accounted for
            num_samples[0] += carryover_num_samples[-1]
            # and truncate as required
            carryover_num_samples = carryover_num_samples[:-1]
        return np.concatenate((carryover_num_samples, num_samples))
    

    def __reconstruct_initial_sweep(self,
                                    previous_chunk: SPECTREChunk,
                                    iq_data: np.ndarray,
                                    sweep_metadata: SweepMetadata) -> Tuple[np.ndarray, SweepMetadata, int]:
        """Reconstruct the initial sweep of the current chunk, using data from the previous chunk."""

        # carryover the final sweep of the previous chunk, and prepend that data to the current chunk data
        carryover_iq_data, carryover_sweep_metadata = self.__get_final_sweep(previous_chunk)
        iq_data = self.__prepend_iq_data(carryover_iq_data,
                                         iq_data)
        
        # if the final sweep of the previous sweep has bled through to the current chunk
        final_sweep_spans_two_chunks = carryover_sweep_metadata.center_frequencies[-1] == sweep_metadata.center_frequencies[0]
        
        center_frequencies = self.__prepend_center_frequencies(carryover_sweep_metadata.center_frequencies,
                                                               sweep_metadata.center_frequencies,
                                                               final_sweep_spans_two_chunks)
        num_samples = self.__prepend_num_samples(carryover_sweep_metadata.num_samples,
                                                 sweep_metadata.num_samples,
                                                 final_sweep_spans_two_chunks)
        
        num_samples_prepended = np.sum(carryover_sweep_metadata.num_samples)
        
        return iq_data, SweepMetadata(center_frequencies, num_samples), num_samples_prepended


    def __correct_timing(self,
                         millisecond_correction: int,
                         num_samples_prepended: int):
        """Correct the start time for this chunk based on the number of samples we prepended reconstructing the initial sweep."""
        elapsed_time = num_samples_prepended * (1 / self.capture_config.get("samp_rate"))

        corrected_datetime = self.chunk_start_datetime + timedelta(milliseconds = millisecond_correction) - timedelta(seconds = float(elapsed_time))
        return corrected_datetime.strftime(DEFAULT_DATETIME_FORMAT), corrected_datetime.microsecond * 1e-3


    def __validate_center_frequencies_ordering(self,
                                               center_frequencies) -> None:
        """Check that the center frequencies are well-ordered in the detached header."""
        min_frequency = np.min(center_frequencies)
        diffs = np.diff(center_frequencies)
        # Extract the expected difference between each step within a sweep. 
        freq_step = self.capture_config.get("freq_step")
        # Validate frequency steps
        for i, diff in enumerate(diffs):
            # steps should either increase by freq_step or drop to the minimum
            if (diff != freq_step) and (center_frequencies[i + 1] != min_frequency):
                raise InvalidSweepMetadataError(f"Unordered center frequencies detected")
    

    def __compute_num_steps_per_sweep(self,
                                      center_frequencies: np.ndarray) -> int:
        """Compute the (ensured constant) number of steps in each sweep."""
        # find the (step) indices corresponding to the minimum frequencies
        min_freq_indices = np.where(center_frequencies == np.min(center_frequencies))[0]
        # then, we evaluate the number of steps that has occured between them via np.diff over the indices
        unique_num_steps_per_sweep = np.unique(np.diff(min_freq_indices))
        # we expect that the difference is always the same, so that the result of np.unique has a single element
        if len(unique_num_steps_per_sweep) != 1:
            raise InvalidSweepMetadataError(("Irregular step count per sweep, "
                                             "expected a consistent number of steps per sweep"))
        return int(unique_num_steps_per_sweep[0])
    

    def __compute_num_full_sweeps(self,
                                  center_frequencies: np.ndarray) -> int:
        """Compute the total number of full sweeps over the chunk.

        Since the number of each samples in each step is variable, we only know a sweep is complete
        when there is a sweep after it. So we can define the total number of *full* sweeps as the number of 
        (freq_max, freq_min) pairs in center_frequencies. It is only at an instance of (freq_max, freq_min) pair 
        in center frequencies that the frequency decreases, so, we can compute the number of full sweeps by 
        counting the numbers of negative values in np.diff(center_frequencies)
        """
        return len(np.where(np.diff(center_frequencies) < 0)[0])


    def __compute_num_max_slices_in_step(self, num_samples: np.ndarray) -> int:
        """Compute the maximum number of slices over all steps, in all sweeps over the chunk."""
        return self.SFT.upper_border_begin(np.max(num_samples))[1]


    def __fill_stepped_dynamic_spectra(self,
                                       stepped_dynamic_spectra: np.ndarray,
                                       iq_data: np.ndarray,
                                       num_samples: np.ndarray,
                                       num_full_sweeps: int,
                                       num_steps_per_sweep: int) -> None:
        """Compute the dynamic spectra for the input IQ samples for each step.
        
        All IQ samples per step were collected at the same center frequency.
        """
        # global_step_index will hold the step index over all sweeps (doesn't reset each sweep)
        # start_sample_index will hold the index of the first sample in the step
        global_step_index, start_sample_index = 0, 0
        for sweep_index in range(num_full_sweeps):
            for step_index in range(num_steps_per_sweep):
                # extract how many samples are in the current step from the metadata
                end_sample_index = start_sample_index + num_samples[global_step_index]
                # compute the number of slices in the current step based on the window we defined on the capture config
                num_slices = self.SFT.upper_border_begin(num_samples[global_step_index])[1]
                # perform a short time fast fourier transform on the step
                complex_spectra = self.SFT.stft(iq_data[start_sample_index:end_sample_index], 
                                           p0=0, 
                                           p1=num_slices)
                # and pack the absolute values into the stepped spectrogram where the step slot is padded to the maximum size for ease of processing later)
                stepped_dynamic_spectra[sweep_index, step_index, :, :num_slices] = np.abs(complex_spectra)
                # reassign the start_sample_index for the next step
                start_sample_index = end_sample_index
                # and increment the global step index
                global_step_index += 1
        

    def __fill_frequencies(self,
                           frequencies: np.ndarray,
                           center_frequencies: np.ndarray,
                           baseband_frequencies: np.ndarray,
                           window_size: int) -> None:
        """Assign physical frequencies to each of the swept spectral components."""
        for i, center_frequency in enumerate(np.unique(center_frequencies)):
            lower_bound = i * window_size
            upper_bound = (i + 1) * window_size
            frequencies[lower_bound:upper_bound] = (baseband_frequencies + center_frequency)
    

    def __fill_times(self,
                     times: np.ndarray,
                     num_samples: np.ndarray,
                     num_full_sweeps: int,
                     num_steps_per_sweep: int) -> None:
        """Assign physical times to each swept spectrum. We use (by convention) the time of the midpoint sample in each sweep."""

        sampling_interval = 1 / self.capture_config.get("samp_rate")
        cumulative_samples = 0
        for sweep_index in range(num_full_sweeps):
            # find the total number of samples across the sweep
            start_step = sweep_index * num_steps_per_sweep
            end_step = (sweep_index + 1) * num_steps_per_sweep
            num_samples_in_sweep = np.sum(num_samples[start_step:end_step])
            
            # compute the midpoint sample in the sweep
            midpoint_sample = cumulative_samples + num_samples_in_sweep // 2

            # update cumulative samples
            cumulative_samples += num_samples_in_sweep

            # assign a physical time to the spectrum for this sweep
            times[sweep_index] = midpoint_sample * sampling_interval
    
    
    def __average_over_steps(self, 
                             stepped_dynamic_spectra: np.ndarray) -> None:
        """Average the spectrums in each step totally in time."""
        return np.nanmean(stepped_dynamic_spectra[..., 1:], axis=-1)
    

    def __stitch_steps(self,
                       stepped_dynamic_spectra: np.ndarray,
                       num_full_sweeps: int) -> np.ndarray:
        """For each full sweep, create a swept spectrum by stitching together the spectrum at each of the steps."""
        return stepped_dynamic_spectra.reshape((num_full_sweeps, -1)).T
    

    def __do_STFFT(self,
                   iq_data: np.ndarray,
                   sweep_metadata: SweepMetadata):
        """Perform a Short Time FFT on the input swept IQ samples."""
        self.__validate_center_frequencies_ordering(sweep_metadata.center_frequencies)

        window_size = len(self.SFT.win)

        num_steps_per_sweep = self.__compute_num_steps_per_sweep(sweep_metadata.center_frequencies)
        num_full_sweeps = self.__compute_num_full_sweeps(sweep_metadata.center_frequencies)
        num_max_slices_in_step = self.__compute_num_max_slices_in_step(sweep_metadata.num_samples)
        
        stepped_dynamic_spectra_shape = (num_full_sweeps, 
                                         num_steps_per_sweep, 
                                         window_size, 
                                         num_max_slices_in_step)
        stepped_dynamic_spectra = np.full(stepped_dynamic_spectra_shape, np.nan)

        
        frequencies = np.empty(num_steps_per_sweep * window_size)
        times = np.empty(num_full_sweeps)

        self.__fill_stepped_dynamic_spectra(stepped_dynamic_spectra,
                                            iq_data,
                                            sweep_metadata.num_samples,
                                            num_full_sweeps,
                                            num_steps_per_sweep)
        
        self.__fill_frequencies(frequencies,
                                sweep_metadata.center_frequencies,
                                self.SFT.f,
                                window_size)
        
        self.__fill_times(times,
                          sweep_metadata.num_samples,
                          num_full_sweeps,
                          num_steps_per_sweep)

        averaged_spectra = self.__average_over_steps(stepped_dynamic_spectra)
        dynamic_spectra = self.__stitch_steps(averaged_spectra,
                                              num_full_sweeps)

        return times, frequencies, dynamic_spectra
    

class HdrChunk(ChunkFile):
    def __init__(self, chunk_parent_path: str, chunk_name: str):
        super().__init__(chunk_parent_path, chunk_name, "hdr")

    def read(self) -> Tuple[int, SweepMetadata]:
        hdr_contents = self._read_file_contents()
        millisecond_correction = self._get_millisecond_correction(hdr_contents)
        center_frequencies = self._get_center_frequencies(hdr_contents)
        num_samples = self._get_num_samples(hdr_contents)
        self._validate_frequencies_and_samples(center_frequencies, 
                                               num_samples)
        
        return millisecond_correction, SweepMetadata(center_frequencies, num_samples)
        

    def _read_file_contents(self) -> np.ndarray:
        with open(self.file_path, "rb") as fh:
            return np.fromfile(fh, dtype=np.float32)


    def _get_millisecond_correction(self, hdr_contents: np.ndarray) -> int:
        ''' Millisecond correction is an integral quantity, but stored in the detached header as a 32-bit float.'''
        millisecond_correction_as_float = float(hdr_contents[0])

        if not millisecond_correction_as_float.is_integer():
            raise TypeError(f"Expected integer value for millisecond correction, but got {millisecond_correction_as_float}")
        
        return int(millisecond_correction_as_float)


    def _get_center_frequencies(self, hdr_contents: np.ndarray) -> np.ndarray:
        ''' 
        Detached header contents are stored in (center_freq_i, num_samples_at_center_freq_i) pairs
        Return only a list of center frequencies, by skipping over file contents in twos.
        '''
        return hdr_contents[1::2]


    def _get_num_samples(self, hdr_contents: np.ndarray) -> np.ndarray:
        ''' 
        Detached header contents are stored in (center_freq_i, num_samples_at_center_freq_i) pairs
        Return only the number of samples at each center frequency, by skipping over file contents in twos.
        Number of samples is an integral quantity, but stored in the detached header as a 32-bit float.
        Types are checked before return.
        '''
        num_samples_as_float = hdr_contents[2::2]
        if not all(num_samples_as_float == num_samples_as_float.astype(int)):
            raise InvalidSweepMetadataError("Number of samples per frequency is expected to describe an integer")
        return num_samples_as_float.astype(int)


    def _validate_frequencies_and_samples(self, center_frequencies: np.ndarray, num_samples: np.ndarray) -> None:
        """Validates that the center frequencies and the number of samples arrays have the same length."""
        if len(center_frequencies) != len(num_samples):
            raise InvalidSweepMetadataError("Center frequencies and number of samples arrays are not the same length")



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod
from typing import Callable, Any, Optional

from spectre_core.receivers import validators
from spectre_core.file_handlers.configs import (
    CaptureConfig,
    validate_against_type_template,
    type_cast_params
)
from spectre_core.exceptions import (
    TemplateNotFoundError,
    ModeNotFoundError,
    SpecificationNotFoundError
)


class BaseReceiver(ABC):
    def __init__(self, name: str, mode: Optional[str] = None):
        self._name = name

        self._capture_methods: dict[str, Callable] = None
        self._set_capture_methods()

        self._type_templates: dict[str, dict[str, Any]] = None
        self._set_type_templates()

        self._validators: dict[str, Callable] = None
        self._set_validators()

        self._specifications: dict[str, Any] = None
        self._set_specifications()

        self.mode = mode



    @abstractmethod
    def _set_capture_methods(self) -> None:
        pass


    @abstractmethod
    def _set_validators(self) -> None:
        pass

    @abstractmethod
    def _set_type_templates(self) -> None:
        pass


    @abstractmethod
    def _set_specifications(self) -> None:
        pass

    
    @property
    def name(self) -> str:
        return self._name
    

    @property
    def capture_methods(self) -> dict[str, Callable]:
        return self._capture_methods
    

    @property
    def validators(self) -> dict[str, Callable]:
        return self._validators


    @property
    def type_templates(self) -> dict[str, dict[str, Any]]:
        return self._type_templates


    @property
    def specifications(self) -> dict[dict, Any]:
        return self._specifications


    @property
    def valid_modes(self) -> None:
        capture_method_modes = list(self.capture_methods.keys())
        validator_modes = list(self.validators.keys())
        type_template_modes = list(self.type_templates.keys())

        if capture_method_modes == validator_modes == type_template_modes:
            return capture_method_modes
        else:
            raise ValueError(f"Mode mismatch for the receiver {self.name}. Could not define valid modes")

    @property
    def mode(self) -> str:
        return self._mode
    

    @mode.setter
    def mode(self, value: Optional[str]) -> None:
        if (value is not None) and value not in self.valid_modes:
            raise ModeNotFoundError((f"{value} is not a defined mode for the receiver {self.name}. "
                                     f"Expected one of {self.valid_modes}"))
        self._mode = value


    @property
    def mode_is_set(self) -> bool:
        return (self._mode is not None)


    @property
    def capture_method(self) -> Callable:
        return self.capture_methods[self.mode]


    @property
    def validator(self) -> Callable:
        return self.validators[self.mode]


    @property
    def type_template(self) -> dict[str, Any]:
        return self._type_templates[self.mode]
    

    def get_specification(self, 
                          specification_key: str) -> Any:
        specification = self.specifications.get(specification_key)
        if specification is None:
            expected_specifications = list(self.specifications.keys())
            raise SpecificationNotFoundError(f"Invalid specification '{specification_key}'. Expected one of {expected_specifications}")
        return specification


    def validate_capture_config(self, 
                                capture_config: CaptureConfig) -> None:
        # validate against the active type template
        validate_against_type_template(capture_config, 
                                       self.type_template, 
                                       ignore_keys=["receiver", "mode", "tag"])
        # validate against receiver-specific constraints
        self.validator(capture_config)



    def start_capture(self, 
                      tag: str) -> None:
        capture_config = self.load_capture_config(tag)
        self.capture_method(capture_config)


    def save_params(self, 
                    params: list[str],  
                    tag: str, 
                    doublecheck_overwrite: bool = True) -> None:
        d = type_cast_params(params, 
                             self.type_template)
        
        validate_against_type_template(d,
                                       self.type_template)
        
        self.save_capture_config(d, 
                                 tag, 
                                 doublecheck_overwrite=doublecheck_overwrite)


    def save_capture_config(self, 
                            d: dict[str, Any],
                            tag: str, 
                            doublecheck_overwrite: bool = True) -> None:
        
        self.validate_capture_config(d)

        d.update({"receiver": self.name, 
                  "mode": self.mode, 
                  "tag": tag})
        
        capture_config = CaptureConfig(tag)
        capture_config.save(d, 
                            doublecheck_overwrite = doublecheck_overwrite)


    def load_capture_config(self, 
                            tag: str) -> CaptureConfig:
        
        capture_config = CaptureConfig(tag)

        if capture_config["receiver"] != self.name:
            raise ValueError(f"Capture config receiver mismatch for tag {tag}. Expected {self.name}, got {capture_config['receiver']}")
        
        if capture_config["mode"] != self.mode:
            raise ValueError(f"Mode mismatch for the tag {tag}. Expected {self.mode}, got {capture_config['mode']}")
        
        self.validate_capture_config(capture_config)
        return capture_config


    def get_create_capture_config_cmd(self, 
                                      tag: str, 
                                      as_string: bool = False) -> str:
        """Get a command which can be used to create a capture config with the SPECTRE CLI."""
        command_as_list = ["spectre", "create", "capture-config", 
                           "--tag", tag, 
                           "--receiver", self.name, 
                           "--mode", self.mode]
        for key, value in self.type_template.items():
            command_as_list.extend(["-p", f"{key}={value.__name__}"])

        return " ".join(command_as_list) if as_string else command_as_list


# optional parent class which provides default templates and validators
class SPECTREReceiver(BaseReceiver):
    def __init__(self, *args, **kwargs):
        self.__set_default_type_templates()
        super().__init__(*args, **kwargs)
    

    @property
    def default_type_templates(self) -> dict[str, dict[str, Any]]:
        return self._default_type_templates
    

    def __set_default_type_templates(self) -> None:
        self._default_type_templates = {
            "fixed": {
                "center_freq": float, # [Hz]
                "bandwidth": float, # [Hz]
                "samp_rate": int, # [Hz]
                "IF_gain": int, # [dB]
                "RF_gain": int, # [dB]
                "chunk_size": int, #  [s]
                "joining_time": int, # [s]
                "time_resolution": float, # [s]
                "frequency_resolution": float, # [Hz]
                "window_type": str, # window type for STFFT
                "window_kwargs": dict, # keyword arguments for window function, must be in order as in scipy documentation.
                "window_size": int, # number of samples in STFFT window
                "STFFT_kwargs": dict, # keyword arguments for the scipy STFFT class
                "chunk_key": str, # tag will map to the chunk with this key
                "event_handler_key": str, # tag will map to event handler with this key during post processing
            },
            "sweep": {
                "min_freq": float, # [Hz]
                "max_freq": float, # [Hz]
                "samples_per_step": int, 
                "freq_step": float, # [Hz]
                "bandwidth": float, # [Hz]
                "samp_rate": int, # [Hz]
                "IF_gain": int, # [dB]
                "RF_gain": int, # [dB]
                "chunk_size": int, #  [s]
                "joining_time": int, # [s]
                "time_resolution": float, # [s]
                "frequency_resolution": float, # [Hz]
                "window_type": str, # window type for STFFT
                "window_kwargs": dict, # keyword arguments for window function, must be in order as in scipy documentation.
                "window_size": int, # number of samples in STFFT window
                "STFFT_kwargs": dict, # keyword arguments for the scipy STFFT class
                "chunk_key": str, # tag will map to the chunk with this key
                "event_handler_key": str, # tag will map to event handler with this key during post processing
            }
        }
    

    def _get_default_type_template(self, 
                                   mode: str) -> dict:
        default_type_template = self.default_type_templates[mode]
        if default_type_template is None:
            raise TemplateNotFoundError(f"No default template found for the mode {mode}")
        return default_type_template
    

    def _default_sweep_validator(self, 
                                 capture_config: CaptureConfig) -> None:
        min_freq = capture_config["min_freq"]
        max_freq = capture_config["max_freq"]
        samples_per_step = capture_config["samples_per_step"]
        freq_step = capture_config["freq_step"]
        bandwidth = capture_config["bandwidth"]
        samp_rate = capture_config["samp_rate"]
        IF_gain = capture_config["IF_gain"]
        RF_gain = capture_config["RF_gain"]
        chunk_size = capture_config["chunk_size"]
        time_resolution = capture_config["time_resolution"]
        window_type = capture_config["window_type"]
        window_kwargs = capture_config["window_kwargs"]
        window_size = capture_config["window_size"]
        STFFT_kwargs = capture_config["STFFT_kwargs"]
        chunk_key = capture_config["chunk_key"]
        event_handler_key = capture_config[ "event_handler_key"]

        validators.center_freq_strictly_positive(min_freq)
        validators.center_freq_strictly_positive(max_freq)
        validators.samp_rate_strictly_positive(samp_rate)
        validators.bandwidth_strictly_positive(bandwidth)
        validators.nyquist_criterion(samp_rate, 
                                              bandwidth)
        validators.chunk_size_strictly_positive(chunk_size)
        validators.time_resolution(time_resolution, 
                                             chunk_size) 
        validators.window(window_type, 
                          window_kwargs, 
                          window_size,
                          chunk_size,
                          samp_rate)
        validators.STFFT_kwargs(STFFT_kwargs)
        validators.chunk_key(chunk_key, "sweep")
        validators.event_handler_key(event_handler_key, "sweep")
        validators.gain_is_negative(IF_gain)
        validators.gain_is_negative(RF_gain)
        validators.num_steps_per_sweep(min_freq, 
                                       max_freq, 
                                       samp_rate, 
                                       freq_step)
        validators.sweep_interval(min_freq, 
                                  max_freq, 
                                  samp_rate, 
                                  freq_step,
                                  samples_per_step,
                                  chunk_size)
        validators.non_overlapping_steps(freq_step, 
                                         samp_rate)
        validators.num_samples_per_step(samples_per_step, 
                                        window_size)

        # if the api latency is defined, raise a warning if the step interval is of the same order
        api_latency = self.specifications.get("api_latency")
        if api_latency:
            validators.step_interval(samples_per_step, 
                                     samp_rate, 
                                     api_latency)
    

    def _default_fixed_validator(self, 
                                 capture_config: CaptureConfig) -> None:
        center_freq = capture_config["center_freq"]
        bandwidth = capture_config["bandwidth"]
        samp_rate = capture_config["samp_rate"]
        IF_gain = capture_config["IF_gain"]
        RF_gain = capture_config["RF_gain"]
        chunk_size = capture_config["chunk_size"]
        time_resolution = capture_config["time_resolution"]
        window_type = capture_config["window_type"]
        window_kwargs = capture_config["window_kwargs"]
        window_size = capture_config["window_size"]
        STFFT_kwargs = capture_config["STFFT_kwargs"]
        chunk_key = capture_config["chunk_key"]
        event_handler_key = capture_config["event_handler_key"]

        validators.center_freq_strictly_positive(center_freq)
        validators.samp_rate_strictly_positive(samp_rate)
        validators.bandwidth_strictly_positive(bandwidth)
        validators.nyquist_criterion(samp_rate, bandwidth)
        validators.chunk_size_strictly_positive(chunk_size)
        validators.time_resolution(time_resolution, chunk_size) 
        validators.window(window_type, 
                          window_kwargs, 
                          window_size,
                          chunk_size,
                          samp_rate)
        validators.STFFT_kwargs(STFFT_kwargs)
        validators.chunk_key(chunk_key, 
                             "fixed")
        validators.event_handler_key(event_handler_key, 
                                     "fixed")
        validators.gain_is_negative(IF_gain)
        validators.gain_is_negative(RF_gain)
    

# parent class for shared methods and attributes of SDRPlay receivers
class SDRPlayReceiver(SPECTREReceiver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _sdrplay_validator(self, 
                           capture_config: CaptureConfig) -> None:
        # RSPduo specific validations in single tuner mode
        center_freq_lower_bound = self.get_specification("center_freq_lower_bound")
        center_freq_upper_bound = self.get_specification("center_freq_upper_bound")
        center_freq = capture_config.get("center_freq")
        min_freq = capture_config.get("min_freq")
        max_freq = capture_config.get("max_freq")

        if center_freq:
            validators.closed_confine_center_freq(center_freq, 
                                                  center_freq_lower_bound, 
                                                  center_freq_upper_bound)
            
        if min_freq:
            validators.closed_confine_center_freq(min_freq, 
                                                  center_freq_lower_bound, 
                                                  center_freq_upper_bound)
        if max_freq:
            validators.closed_confine_center_freq(max_freq, 
                                                  center_freq_lower_bound, 
                                                  center_freq_upper_bound)

        validators.closed_confine_samp_rate(capture_config["samp_rate"], 
                                            self.get_specification("samp_rate_lower_bound"), 
                                            self.get_specification("samp_rate_upper_bound"))


        validators.closed_confine_bandwidth(capture_config["bandwidth"], 
                                            self.get_specification("bandwidth_lower_bound"), 
                                            self.get_specification("bandwidth_upper_bound"))

        validators.closed_upper_bound_IF_gain(capture_config["IF_gain"], 
                                              self.get_specification("IF_gain_upper_bound"))
        
        validators.closed_upper_bound_RF_gain(capture_config["RF_gain"], 
                                              self.get_specification("RF_gain_upper_bound"))



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os

SPECTRE_DATA_DIR_PATH = os.environ.get("SPECTRE_DATA_DIR_PATH")
if SPECTRE_DATA_DIR_PATH is None:
    raise ValueError("The environment variable SPECTRE_DATA_DIR_PATH has not been set")

CHUNKS_DIR_PATH = os.environ.get("SPECTRE_CHUNKS_DIR_PATH", 
                                 os.path.join(SPECTRE_DATA_DIR_PATH, 'chunks'))
os.makedirs(CHUNKS_DIR_PATH, 
            exist_ok=True)

LOGS_DIR_PATH = os.environ.get("SPECTRE_LOGS_DIR_PATH",
                               os.path.join(SPECTRE_DATA_DIR_PATH, 'logs'))
os.makedirs(LOGS_DIR_PATH, 
            exist_ok=True)

JSON_CONFIGS_DIR_PATH = os.environ.get("SPECTRE_JSON_CONFIGS_DIR_PATH",
                                       os.path.join(SPECTRE_DATA_DIR_PATH, "configs"))
os.makedirs(JSON_CONFIGS_DIR_PATH, 
            exist_ok=True)

DEFAULT_TIME_FORMAT = "%H:%M:%S"
DEFAULT_DATE_FORMAT = "%Y-%m-%d"
DEFAULT_DATETIME_FORMAT = f"{DEFAULT_DATE_FORMAT}T{DEFAULT_TIME_FORMAT}"

CALLISTO_INSTRUMENT_CODES = [
    "ALASKA-ANCHORAGE",
    "ALASKA-COHOE",
    "ALASKA-HAARP",
    "ALGERIA-CRAAG",
    "ALMATY",
    "AUSTRIA-Krumbach",
    "AUSTRIA-MICHELBACH",
    "AUSTRIA-OE3FLB",
    "AUSTRIA-UNIGRAZ",
    "Australia-ASSA",
    "BIR",
    "Croatia-Visnjan",
    "DENMARK",
    "EGYPT-Alexandria",
    "EGYPT-SpaceAgency",
    "FINLAND-Siuntio",
    "Finland-Kempele",
    "GERMANY-DLR",
    "GLASGOW",
    "GREENLAND",
    "HUMAIN",
    "HURBANOVO",
    "INDIA-GAURI",
    "INDIA-OOTY",
    "INDIA-UDAIPUR",
    "JAPAN-IBARAKI",
    "KASI",
    "MEXART",
    "MEXICO-FCFM-UANL",
    "MEXICO-LANCE-A",
    "MEXICO-LANCE-B",
    "MONGOLIA-UB",
    "MRO",
    "MRT3",
    "Malaysia-Banting",
    "NORWAY-EGERSUND",
    "NORWAY-NY-AALESUND",
    "NORWAY-RANDABERG",
    "POLAND-Grotniki",
    "ROMANIA",
    "ROSWELL-NM",
    "SPAIN-PERALEJOS",
    "SSRT",
    "SWISS-HB9SCT",
    "SWISS-HEITERSWIL",
    "SWISS-IRSOL",
    "SWISS-Landschlacht",
    "SWISS-MUHEN",
    "TRIEST",
    "TURKEY",
    "UNAM",
    "URUGUAY",
    "USA-BOSTON",
]

def _get_date_based_dir_path(base_dir: str, year: int = None, 
                             month: int = None, day: int = None
) -> str:
    if day and not (year and month):
        raise ValueError("A day requires both a month and a year")
    if month and not year:
        raise ValueError("A month requires a year")
    
    date_dir_components = []
    if year:
        date_dir_components.append(f"{year:04}")
    if month:
        date_dir_components.append(f"{month:02}")
    if day:
        date_dir_components.append(f"{day:02}")
    
    return os.path.join(base_dir, *date_dir_components)


def get_chunks_dir_path(year: int = None, 
                        month: int = None, 
                        day: int = None
) -> str:
    return _get_date_based_dir_path(CHUNKS_DIR_PATH, year, month, day)


def get_logs_dir_path(year: int = None, 
                      month: int = None, 
                      day: int = None
) -> str:
    return _get_date_based_dir_path(LOGS_DIR_PATH, year, month, day)


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from abc import ABC, abstractmethod
from typing import Any, Optional
from warnings import warn

class BaseFileHandler(ABC):
    def __init__(self, 
                 parent_path: str, 
                 base_file_name: str, 
                 extension: Optional[str] = None):
        self._parent_path = parent_path
        self._base_file_name = base_file_name
        self._extension = extension

        
    @abstractmethod
    def read(self) -> Any:
        pass
 

    @property
    def parent_path(self) -> str:
        return self._parent_path
    

    @property
    def base_file_name(self) -> str:
        return self._base_file_name
    

    @property
    def extension(self) -> Optional[str]:
        return self._extension
    

    @property
    def file_name(self) -> str:
        return self._base_file_name if (self._extension is None) else f"{self._base_file_name}.{self._extension}"
    

    @property
    def file_path(self) -> str:
        return os.path.join(self._parent_path, self.file_name)
    
    
    @property
    def exists(self) -> bool:
        return os.path.exists(self.file_path) 


    def make_parent_path(self) -> None:
        os.makedirs(self.parent_path, exist_ok=True) 
    

    def delete(self, 
               doublecheck_delete = True) -> None:
        if not self.exists:
            warn(f"{self.file_path} does not exist. No deletion taking place")
            return
        else:
            if doublecheck_delete:
                self.doublecheck_delete()
            os.remove(self.file_path)
    

    def cat(self) -> None:
        print(self.read())


    def _doublecheck_action(self, 
                            action_message: str) -> None:
        proceed_with_action = False
        while not proceed_with_action:
            user_input = input(f"{action_message} [y/n]: ").strip().lower()
            if user_input == "y":
                proceed_with_action = True
            elif user_input == "n":
                print("Operation cancelled by the user")
                raise exit(1)
            else:
                print(f"Please enter one of [y/n], received {user_input}")
                proceed_with_action = False


    def doublecheck_overwrite(self) -> None:
        self._doublecheck_action(action_message=f"The file '{self.file_path}' already exists. Overwrite?")


    def doublecheck_delete(self) -> None:
        self._doublecheck_action(action_message=f"Are you sure you would like to delete '{self.file_path}'?")


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later


from logging import getLogger
_LOGGER = getLogger(__name__)

import os
import logging
from typing import Callable, Optional
import warnings
from collections import OrderedDict
from datetime import datetime

from spectre_core.file_handlers.text import TextHandler
from spectre_core.cfg import (
    LOGS_DIR_PATH, 
    DEFAULT_DATETIME_FORMAT, 
    get_logs_dir_path
)

PROCESS_TYPES = [
    "USER", 
    "WORKER"
]


def validate_process_type(process_type: str
) -> None:
    if process_type not in PROCESS_TYPES:
        raise ValueError(f"Invalid process type: {process_type}. Expected one of {PROCESS_TYPES}")


class LogHandler(TextHandler):
    def __init__(self, 
                 datetime_stamp: str, 
                 pid: str, 
                 process_type: str):
        self._datetime_stamp = datetime_stamp
        self._pid = pid
        validate_process_type(process_type)
        self._process_type = process_type

        dt = datetime.strptime(datetime_stamp, DEFAULT_DATETIME_FORMAT)
        date_dir = os.path.join(dt.strftime("%Y"), dt.strftime("%m"), dt.strftime("%d"))
        parent_path = os.path.join(LOGS_DIR_PATH, date_dir)
        base_file_name = f"{datetime_stamp}_{pid}_{process_type}"

        super().__init__(parent_path, base_file_name, extension = "log")
    

    @property
    def datetime_stamp(self) -> str:
        return self._datetime_stamp
    

    @property
    def pid(self) -> str:
        return self._pid
    

    @property
    def process_type(self) -> str:
        return self._process_type


class LogHandlers:
    def __init__(self, 
                 process_type: Optional[str] = None, 
                 year: Optional[int] = None, 
                 month: Optional[int] = None, 
                 day: Optional[int] = None):
        self._log_handler_map: dict[str, LogHandler] = OrderedDict()
        self._process_type = process_type
        self.set_date(year, month, day)


    @property
    def process_type(self) -> str:
        return self._process_type
    

    @property
    def year(self) -> Optional[int]:
        return self._year


    @property 
    def month(self) -> Optional[int]:
        return self._month
    

    @property
    def day(self) -> Optional[int]:
        return self._day


    @property
    def logs_dir_path(self) -> str:
        return get_logs_dir_path(self.year, self.month, self.day)


    @property 
    def log_handler_map(self) -> dict[str, LogHandler]:
        if not self._log_handler_map: # check for empty dictionary
            self._update_log_handler_map()
        return self._log_handler_map
        

    @property
    def log_handler_list(self) -> list[LogHandler]:
        return list(self.log_handler_map.values())


    @property
    def num_logs(self) -> int:
        return len(self.log_handler_list) 


    @property
    def file_names(self) -> list[str]:
        return list(self.log_handler_map.keys())


    def set_date(self, 
                 year: Optional[int],
                 month: Optional[int],
                 day: Optional[int]) -> None:
        self._year = year
        self._month = month
        self._day = day
        self._update_log_handler_map()


    def _update_log_handler_map(self) -> None:
        log_files = [f for (_, _, files) in os.walk(self.logs_dir_path) for f in files]

        if not log_files:
            warning_message = "No logs found, setting log map to an empty dictionary"
            _LOGGER.warning(warning_message)
            warnings.warn(warning_message)
            return

        for log_file in log_files:
            file_name, _ = os.path.splitext(log_file)
            log_start_time, pid, process_type = file_name.split("_")

            if self.process_type and process_type != self.process_type:
                continue

            self._log_handler_map[file_name] = LogHandler(log_start_time, pid, process_type)

        self._log_handler_map = OrderedDict(sorted(self._log_handler_map.items()))


    def update(self) -> None:
        """Public alias for setting log handler map"""
        self._update_log_handler_map()


    def __iter__(self):
        yield from self.log_handler_list


    def get_log_handler_from_file_name(self, 
                                       file_name: str) -> LogHandler:
        # auto strip the extension if present
        file_name, _ = os.path.splitext(file_name)
        try:
            return self.log_handler_map[file_name]
        except KeyError:
            raise FileNotFoundError(f"Log handler for file name '{file_name}' not found in log map")


    def get_log_handler_from_pid(self, 
                                 pid: str) -> LogHandler:
        for log_handler in self.log_handler_list:
            if log_handler.pid == pid:
                return log_handler
        raise FileNotFoundError(f"Log handler for PID '{pid}' not found in log map")


def configure_root_logger(process_type: str, 
                          level: int = logging.INFO
) -> LogHandler:
    system_datetime = datetime.now()
    datetime_stamp = system_datetime.strftime(DEFAULT_DATETIME_FORMAT)
    pid = os.getpid()
    log_handler = LogHandler(datetime_stamp, pid, process_type)
    log_handler.make_parent_path()

    # configure the root logger
    logger = logging.getLogger()
    logger.setLevel(level)
    # Remove any existing handlers to avoid duplicate logs
    for handler in logger.handlers:
        logger.removeHandler(handler)
    # Set up file handler with specific filename
    file_handler = logging.FileHandler(log_handler.file_path)
    file_handler.setLevel(level)
    formatter = logging.Formatter("[%(asctime)s] [%(levelname)8s] --- %(message)s (%(name)s:%(lineno)s)")
    file_handler.setFormatter(formatter)
    # and add it to the root logger
    logger.addHandler(file_handler)

    return log_handler

# Logger must be passed in to preserve context of the service function
def log_call(logger: logging.Logger
) -> Callable:
    def decorator(func: Callable) -> Callable:
        def wrapper(*args, **kwargs):
            try:
                logger.info(f"Calling the function: {func.__name__}")
                return func(*args, **kwargs)
            except Exception as e:
                logger.error(f"An error occurred while calling the function: {func.__name__}",
                              exc_info=True)
                raise
        return wrapper
    return decorator


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime
from abc import abstractmethod
from typing import Optional

from scipy.signal import ShortTimeFFT, get_window

from spectre_core.file_handlers.base import BaseFileHandler
from spectre_core.cfg import get_chunks_dir_path
from spectre_core.file_handlers.configs import CaptureConfig
from spectre_core.spectrograms.spectrogram import Spectrogram
from spectre_core.cfg import DEFAULT_DATETIME_FORMAT
from spectre_core.exceptions import ChunkFileNotFoundError

class ChunkFile(BaseFileHandler):
    def __init__(self, 
                 chunk_parent_path: str, 
                 chunk_name: str, 
                 extension: str, 
                 **kwargs):
        self._chunk_start_time, self._tag = chunk_name.split("_")
        self._chunk_start_datetime: Optional[datetime] = None
        super().__init__(chunk_parent_path, 
                         chunk_name, 
                         extension = extension, 
                         **kwargs)


    @property
    def chunk_start_time(self) -> str:
        return self._chunk_start_time
    

    @property
    def chunk_start_datetime(self) -> datetime:
        if self._chunk_start_datetime is None:
            self._chunk_start_datetime = datetime.strptime(self.chunk_start_time, DEFAULT_DATETIME_FORMAT)
        return self._chunk_start_datetime
    

    @property
    def tag(self) -> str:
        return self._tag
    


class BaseChunk:
    def __init__(self, 
                 chunk_start_time: str, 
                 tag: str):
        self._chunk_start_time: str = chunk_start_time
        self._tag: str = tag
        self._chunk_files: dict[str, ChunkFile] = {}
        self._chunk_start_datetime: Optional[datetime] = None
        self.chunk_parent_path: str = get_chunks_dir_path(year = self.chunk_start_datetime.year,
                                                          month = self.chunk_start_datetime.month,
                                                          day = self.chunk_start_datetime.day)
        self._chunk_name: str = f"{self.chunk_start_time}_{self.tag}"


    @property
    def chunk_start_time(self) -> str:
        return self._chunk_start_time
    

    @property
    def chunk_start_datetime(self) -> datetime:
        if self._chunk_start_datetime is None:
            self._chunk_start_datetime = datetime.strptime(self.chunk_start_time, DEFAULT_DATETIME_FORMAT)
        return self._chunk_start_datetime
    

    @property
    def tag(self) -> str:
        return self._tag


    @property
    def chunk_name(self) -> str:
        return f"{self._chunk_start_time}_{self._tag}"
    

    @property
    def extensions(self) -> list[str]:
        return list(self._chunk_files.keys())
    
    
    def add_file(self, chunk_file: ChunkFile) -> None:
        self._chunk_files[chunk_file.extension] = chunk_file
    

    def get_file(self, extension: str) -> ChunkFile:
        try:
            return self._chunk_files[extension]
        except KeyError:
            raise ChunkFileNotFoundError(f"No chunk file found with extension '{extension}'")


    def read_file(self, extension: str):
        chunk_file = self.get_file(extension)
        return chunk_file.read()


    def delete_file(self, extension: str, **kwargs):
        chunk_file = self.get_file(extension)
        try:
            chunk_file.delete(**kwargs)
        except FileNotFoundError as e:
            raise ChunkFileNotFoundError(str(e))


    def has_file(self, extension: str) -> bool:
        try:
            chunk_file = self.get_file(extension)
            return chunk_file.exists
        except ChunkFileNotFoundError:
            return False


class SPECTREChunk(BaseChunk):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._capture_config = CaptureConfig(self._tag)
        self._SFT = None # cache


    @abstractmethod
    def build_spectrogram(self) -> Spectrogram:
        """Create a spectrogram object derived from chunk files for this chunk."""
        pass


    @property
    def capture_config(self) -> CaptureConfig:
        return self._capture_config
    

    @property
    def SFT(self) -> ShortTimeFFT:
        if self._SFT is None:
            self._SFT = self.__get_SFT_instance()
        return self._SFT
    

    def __get_SFT_instance(self) -> ShortTimeFFT:
        window_type = self.capture_config.get("window_type")
        window_params = self.capture_config.get("window_kwargs").values()
        window_size = self.capture_config.get("window_size")
        window = get_window((window_type, 
                             *window_params), 
                             window_size)
        samp_rate = self.capture_config.get("samp_rate")
        return ShortTimeFFT(window, 
                            fs=samp_rate, 
                            fft_mode="centered", 
                            **self.capture_config.get("STFFT_kwargs"))